"""Information about this package.

See custombot.py for information about Bot class.
"""

__author__ = "Davide Testa"
__email__ = "davide@davte.it"
__credits__ = "Marco Origlia"
__license__ = "GNU General Public License v3.0"
__version__ = "1.5.9"
__maintainer__ = "Davide Testa"
__contact__ = "t.me/davte"

from davtelepot.custombot import Bot
